import Foundation

//
//  MeshBuilder.swift
//  MetalRenderer
//
//  Created by Caroline on 20/2/20.
//  Copyright © 2020 Razeware. All rights reserved.
//

import MetalKit

enum PrimitiveType {
  case rect
  case cube
}

enum CubeSize {
  case front, back, top, bottom, left, right
}

class Primitive {
  struct Vertex {
    var position: float3
  }
  
  struct Mesh {
    let vertexBuffer: MTLBuffer
    let submeshes: [Submesh]
  }
  
  struct Submesh {
    let indexBuffer: MTLBuffer
    let indexCount: Int
    let indexType: MTLIndexType
    var color: float3 = [1, 0, 0]
  }
  
  let mesh: Mesh
  let boundingBox: MDLAxisAlignedBoundingBox
  
  init(_ primitiveType: PrimitiveType,
       size: float3) {
    
    switch primitiveType {
    case .rect:
      mesh = Self.buildRect(size: size)
      
    case .cube:
      mesh = Self.buildCube(size: size)
    }
    self.boundingBox = Self.boundingBox(size: size)
  }
  
  
  static func boundingBox(size: float3) -> MDLAxisAlignedBoundingBox {
    let halfSize = Self.half(size)
    let minBounds = float3(-halfSize.x, -halfSize.y, -halfSize.z)
    let maxBounds = float3(halfSize.x, halfSize.y, halfSize.z)
    return MDLAxisAlignedBoundingBox(maxBounds: maxBounds,
                                     minBounds: minBounds)
  }
  
  static func half(_ size: float3) -> float3 {
    [size.x * 0.5, size.y * 0.5, size.z * 0.5]
  }
  
  static func buildRect(size: float3) -> Mesh {
    var vertices: [Vertex] {
      let x = half(size).x
      let y = half(size).y
      let z = half(size).z
      let vertices =
        [
          Vertex(position: [-x,  y, z]),
          Vertex(position: [ x,  y, z]),
          Vertex(position: [ x, -y, z]),
          Vertex(position: [-x, -y, z])
      ]
      for vertex in vertices {
        print(vertex.position)
      }
      return vertices
    }
    
        let indices: [UInt16] = [
          0, 1, 2,
          0, 2, 3
        ]
    let vertexBuffer = Renderer.device.makeBuffer(bytes: vertices,
                                                  length: MemoryLayout<Vertex>.stride * vertices.count,
                                                  options: [])!
    let indexBuffer = Renderer.device.makeBuffer(bytes: indices,
                                                  length: MemoryLayout<UInt16>.stride * indices.count,
                                                  options: [])!
    var submesh = Submesh(indexBuffer: indexBuffer,
                          indexCount: indices.count,
                          indexType: .uint16)
    submesh.color = [1, 0, 1]
    let mesh = Mesh(vertexBuffer: vertexBuffer,
                    submeshes: [submesh])
    return mesh
  }
  
  static func buildCube(size: float3) -> Mesh {
    var vertices: [Vertex] {
      let x = half(size).x
      let y = half(size).y
      let z = half(size).z
      let vertices: [Vertex] = [
        Vertex(position: [-x, y, z]),
        Vertex(position: [-x, -y, z]),
        Vertex(position: [x, -y, z]),
        Vertex(position: [x, y, z]),
        Vertex(position: [-x, y, -z]),
        Vertex(position: [x, y, -z]),
        Vertex(position: [-x, -y, -z]),
        Vertex(position: [x, -y, -z])
      ]
      return vertices
    }
    let indices: [[UInt16]] = [
      [0, 1, 2, 0, 2, 3],  // Front
      [5, 7, 6, 5, 6, 4],  // Back
      [4, 6, 1, 4, 1, 0],  // Left
      [3, 2, 7, 3, 7, 5],  // Right
      [4, 0, 3, 4, 3, 5],  // Top
      [1, 6, 7, 1, 7, 2]   // Bottom
    ]
    
    let colors: [float3] = [
      [1, 0, 0],   //Front red
      [0, 1, 0],   //Back green
      [0, 0, 1],   //Left blue
      [1, 1, 0],   //Right yellow
      [0, 1, 1],   //Top cyan
      [1, 0, 1]    //Bottom magenta
    ]
    
    
    var submeshes: [Submesh] = []
    for (index, group) in indices.enumerated() {
      let indexBuffer = Renderer.device.makeBuffer(bytes: group,
                                                   length: MemoryLayout<UInt16>.stride * group.count,
                                                   options: [])!
      
      var submesh = Submesh(indexBuffer: indexBuffer,
                            indexCount: group.count,
                            indexType: .uint16)
      submesh.color = colors[index]
      submeshes.append(submesh)
    }
    
    let vertexBuffer = Renderer.device.makeBuffer(bytes: vertices,
                                                  length: MemoryLayout<Vertex>.stride * vertices.count,
                                                  options: [])!
    let mesh = Mesh(vertexBuffer: vertexBuffer, submeshes: submeshes)
    return mesh
  }
  
}
