
import MetalKit

class Renderer: NSObject {
  
  static var device: MTLDevice!
  let commandQueue: MTLCommandQueue
  var pipelineState: MTLRenderPipelineState!
  var depthStencilState: MTLDepthStencilState!
  
  let camera = Camera()
  
  var renderPipelineState: MTLRenderPipelineState!
  var transparencyEnabled = false
  var blendingEnabled = false
  
  var angleX = Float(0.0)
  var angleY = Float.pi / 3.0
  var angleZ = Float.pi / 7.0
  
  lazy var model: Primitive = {
    return Primitive(.cube, size: [1, 1, 1])
  }()
  
  public init(metalView: MTKView) {
    guard let device = MTLCreateSystemDefaultDevice() else {
      fatalError("GPU not available")
    }
    metalView.device = device
    Renderer.device = device
    commandQueue = device.makeCommandQueue()!
    super.init()
    metalView.clearColor = MTLClearColor(red: 1,
                                         green: 1,
                                         blue: 0.8,
                                         alpha: 1)
    metalView.depthStencilPixelFormat = .depth32Float
    
    
    pipelineState = Self.buildPipelineState()
    buildDepthStencilState()
    
    camera.transform.position = [0, 2, -1]
  }
  
  static func buildPipelineState() -> MTLRenderPipelineState {
    let pipelineState: MTLRenderPipelineState
    let library = Self.device.makeDefaultLibrary()!
    let vertexFunction = library.makeFunction(name: "vertex_main")
    let fragmentFunction = library.makeFunction(name: "fragment_main")
    let descriptor = MTLRenderPipelineDescriptor()
    descriptor.colorAttachments[0].pixelFormat = .bgra8Unorm
    descriptor.depthAttachmentPixelFormat = .depth32Float
    descriptor.vertexFunction = vertexFunction
    descriptor.fragmentFunction = fragmentFunction
    do {
      pipelineState = try Self.device.makeRenderPipelineState(descriptor: descriptor)
    } catch let error {
      fatalError(error.localizedDescription)
    }
    return pipelineState
  }
  
  func buildDepthStencilState() {
    let descriptor = MTLDepthStencilDescriptor()
    descriptor.depthCompareFunction = .less
    descriptor.isDepthWriteEnabled = true
    depthStencilState = Self.device.makeDepthStencilState(descriptor: descriptor)
  }
  
  public func zoomUsing(delta: CGFloat) {
    let sensitivity = Float(0.1)
    let rotation = camera.transform.rotation
    let dx = Float(delta) * sensitivity * sin(rotation.y)
    let dz = Float(delta) * sensitivity * cos(rotation.y)
    camera.transform.position.x -= dx
    camera.transform.position.z += dz
  }
  
  public func rotateUsing(translation: NSPoint) {
    let sensitivity: Float = 0.01
    camera.transform.rotation.x -= Float(translation.y) * sensitivity
    camera.transform.rotation.y += Float(translation.x) * sensitivity
  }
}

extension Renderer: MTKViewDelegate {
  public func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
    camera.aspect = Float(size.width/size.height)
  }
  
  public func draw(in view: MTKView) {
    guard let descriptor = view.currentRenderPassDescriptor,
      let commandBuffer = commandQueue.makeCommandBuffer(),
      let renderEncoder = commandBuffer.makeRenderCommandEncoder(descriptor: descriptor),
      let drawable = view.currentDrawable else {
        return
    }
    renderEncoder.setCullMode(.back)
    renderEncoder.setDepthStencilState(depthStencilState)
    
    renderEncoder.setRenderPipelineState(pipelineState)
    
    var matrix = camera.projectionMatrix * camera.viewMatrix * calculateMatrix()
    renderEncoder.setVertexBytes(&matrix, length: MemoryLayout<float4x4>.stride, index: 11)
    
    renderEncoder.setVertexBuffer(model.mesh.vertexBuffer, offset: 0, index: 0)

    renderEncoder.setTriangleFillMode(.fill)
    var isWireframe = false
    renderEncoder.setFragmentBytes(&isWireframe, length: MemoryLayout<Bool>.stride, index: 1)

    for submesh in model.mesh.submeshes {
      var color = submesh.color
      renderEncoder.setFragmentBytes(&color, length: MemoryLayout<float3>.stride, index: 0)
      renderEncoder.drawIndexedPrimitives(type: .triangle,
                                          indexCount: submesh.indexCount,
                                          indexType: submesh.indexType,
                                          indexBuffer: submesh.indexBuffer,
                                          indexBufferOffset: 0)
    }

    // draw wireframe
    renderEncoder.setTriangleFillMode(.lines)
    isWireframe = true
    renderEncoder.setFragmentBytes(&isWireframe, length: MemoryLayout<Bool>.stride, index: 1)
    for submesh in model.mesh.submeshes {
      var color: float3 = [0, 0, 0]
      renderEncoder.setFragmentBytes(&color, length: MemoryLayout<float3>.stride, index: 0)
      renderEncoder.drawIndexedPrimitives(type: .triangle,
                                          indexCount: submesh.indexCount,
                                          indexType: submesh.indexType,
                                          indexBuffer: submesh.indexBuffer,
                                          indexBufferOffset: 0)
    }

    renderEncoder.endEncoding()
    commandBuffer.present(drawable)
    commandBuffer.commit()
  }
  
  func calculateMatrix() -> float4x4 {
    angleX = angleX + 0.001
    angleY = angleY + 0.003
    angleZ = angleZ + 0.005
    
    var rotateX = float4x4.identity()
    var rotateY = float4x4.identity()
    var rotateZ = float4x4.identity()
    // on x
    rotateX.columns.0 = [1, 0, 0, 0]
    rotateX.columns.1 = [0.0, cos(angleX), -sin(angleX), 0]
    rotateX.columns.2 = [0.0, sin(angleX), cos(angleX), 0]
    // on y
    rotateY.columns.0 = [cos(angleY), 0, sin(angleY), 0]
    rotateY.columns.1 = [0, 1, 0, 0]
    rotateY.columns.2 = [-sin(angleY), 0, cos(angleY), 0]
    // on z
    rotateZ.columns.0 = [cos(angleZ), -sin(angleZ), 0, 0]
    rotateZ.columns.1 = [sin(angleZ), cos(angleZ), 0, 0]
    rotateZ.columns.2 = [0, 0, 1, 0]
    
    var translate = float4x4.identity()
    translate.columns.3 = [0, 2, 4, 1]
    return translate * rotateZ * rotateY * rotateX
  }
}
